package com.seitenbau.junit;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.junit.Ignore;
import org.junit.Test;
import org.junit.internal.runners.ClassRoadie;
import org.junit.internal.runners.CompositeRunner;
import org.junit.internal.runners.InitializationError;
import org.junit.internal.runners.JUnit4ClassRunner;
import org.junit.internal.runners.TestClass;
import org.junit.internal.runners.TestMethod;
import org.junit.runner.Description;
import org.junit.runner.manipulation.Filter;
import org.junit.runner.manipulation.Filterable;
import org.junit.runner.manipulation.NoTestsRemainException;
import org.junit.runner.notification.RunNotifier;

import com.seitenbau.junit.annotation.ParameterizedTest;
import com.seitenbau.junit.annotation.ParameterizedTestListener;
import com.seitenbau.junit.annotation.PrepareFixture;

/**
 * Implementation of Data-Driven Tests
 *
 * It's a simple Implementation for writing Data-Driven Tests with
 * JUnit.
 *
 * Activate on JUnit Test Class via the annotation :
 * <code>@RunWith(DataDrivenTestRunner.class)</code>
 *
 * The following Annotations are also supported:<br>
 * <b>on test class level:</b><br>
 * <li><code>@DataDrivenTestListener( { ? extends IDataDrivenTestListener } )</code></li>
 *
 * <br>
 * <b>on method level:</b><br>
 * <li><code>@Test</code> normal JUnit Test.</li>
 *       <li><code>@Ignore</code> ignores the test.</li>
 *         <li><code>@DataDrivenTest()</code> marks the method as Data Driven Test.
 *                   Dependent on the passed provider additional
 *                   Annotations are needed. </li>
 *                   <li><code></code></li>
 *
 * <br>
 * <b>on field level:</b><br>
 * <li><code>@PrepareFixture</code> marks one method that is called directly
 *                 after creating the test, but before the
 *                 DataDrivenTestListener s were called.</li>
 *                 <li><code>@DataDrivenTestField</code> marks field to get injected by
 *                      'something'. Without parameter this is a
 *                      ITestContext instance </li>
 */
public class ParameterizedTestRunner extends CompositeRunner implements Filterable
{

  public ParameterizedTestRunner(Class<?> testClass) throws Exception
  {
    super(testClass.getCanonicalName());
    init(testClass);
  }

  protected void init(Class<?> testClass) throws Exception
  {
    TestClass theTestClass = new TestClass(testClass);

    List<Method> methods = theTestClass
        .getAnnotatedMethods(ParameterizedTest.class);
    for (Method method : methods)
    {
      // IF junit:@Ignore then ignore the Test
      if (method.getAnnotation(Ignore.class) == null)
      {
        DataDrivenTestMethodRunner runner = new DataDrivenTestMethodRunner(
            theTestClass, method);
        add(runner);
      }
    }

    // Still do the normal junit:@Test Tests
    if (!theTestClass.getAnnotatedMethods(Test.class).isEmpty())
    {
      add(new NormalJUnit4ClassRunner(theTestClass, getDescription()));
    }

    if (getRunners().isEmpty())
    {
      throw new InitializationError(new Exception("No runnable methods"));
    }
  }

}

/**
 * Helper to execute 'normal' JUnit Tests inside the test class.
 */
class NormalJUnit4ClassRunner extends JUnit4ClassRunner
{
  public NormalJUnit4ClassRunner(TestClass testClass,
      Description rootDescription) throws InitializationError
  {
    super(testClass.getJavaClass());
  }

  @Override
  public Description getDescription()
  {
    Description spec = Description.createSuiteDescription(
        "Normal Test Methods [" + getName() + "]", classAnnotations());
    List<Method> testMethods = getTestMethods();
    for (Method method : testMethods)
    {

      spec.addChild(methodDescription(method));
    }
    return spec;
  }

}

/**
 * Modifier JUnit4ClassRunnter for running the DataDriven Tests
 */
class DataDrivenTestMethodRunner extends JUnit4ClassRunner
{

  private TestClass myTestClass;

  private List<ITestData> myTestData = new ArrayList<ITestData>();

  private Method myTestMethod;

  /**
   * Hack, for passing the ITestData instance from run() to
   * createTest()
   */
  private ITestData currentTestData;

  private IDataTestProvider myProvider;

  private Method myCreateFixureMethod;

  private ArrayList<IParameterizedTestListener> myTestExecution;

  public DataDrivenTestMethodRunner(TestClass testClass, Method method)
      throws Exception
  {
    super(testClass.getJavaClass());
    init(testClass, method);

    loadTestExecutions(testClass.getJavaClass().getAnnotation(
        ParameterizedTestListener.class), testClass.getJavaClass(), method);
  }

  private void loadTestExecutions(ParameterizedTestListener annotation,
      Class<?> testClazz, Method method) throws Exception
  {
    myTestExecution = new ArrayList<IParameterizedTestListener>();
    if (annotation == null)
    {
      return;
    }
    Class<? extends IParameterizedTestListener>[] listenersClasses = annotation
        .value();
    for (int i = 0; i < listenersClasses.length; i++)
    {
      Class<? extends IParameterizedTestListener> listenerClass = listenersClasses[i];
      IParameterizedTestListener obj = listenerClass.newInstance();
      IParameterizedTestListener listener = (IParameterizedTestListener) obj;
      myTestExecution.add(listener);
      listener.init(testClazz, method);
    }
  }

  public ArrayList<IParameterizedTestListener> getTestExecutions()
  {
    return myTestExecution;
  }

  protected void init(TestClass testClass, Method method)
  {
    myTestClass = testClass;
    myTestMethod = method;

    List<Method> createFixuresMethods = myTestClass
        .getAnnotatedMethods(PrepareFixture.class);
    myCreateFixureMethod = null;
    if (createFixuresMethods != null)
    {
      if (createFixuresMethods.size() > 1)
      {
        throw new IllegalArgumentException("To many @"
            + PrepareFixture.class.getSimpleName() + " methods");
      }
      else if (createFixuresMethods.size() == 1)
      {
        myCreateFixureMethod = createFixuresMethods.get(0);
      }
    }
    Class<? extends IDataTestProvider> providerClass = method.getAnnotation(
        ParameterizedTest.class).value();
    try
    {
      myProvider = providerClass.newInstance();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
    myProvider.init(testClass.getJavaClass(), method);
    myTestData.addAll(myProvider.buildTestData());
  }

  @Override
  public void filter(Filter filter) throws NoTestsRemainException
  {
    for (Iterator<ITestData> iter = myTestData.iterator(); iter.hasNext();)
    {
      if (!filter.shouldRun(iter.next().getDescription()))
      {
        iter.remove();
      }
    }
  }

  @Override
  public Description getDescription()
  {
    Description root = Description.createSuiteDescription(myTestMethod
        .getName()
        + " [" + myProvider.getDescriptionPostText() + "]");
    for (ITestData testData : myTestData)
    {
      Description desc = testData.getDescription();
      root.addChild(desc);
    }
    return root;
  }

  /** needed to do nothing */
  @Override
  protected void validate() throws InitializationError
  {
  }

  @Override
  public void run(final RunNotifier notifier)
  {
    for (Iterator<ITestData> iter = myTestData.iterator(); iter.hasNext();)
    {
      currentTestData = iter.next();
      new ClassRoadie(notifier, myTestClass, currentTestData.getDescription(),
          new Runnable()
          {
            public void run()
            {
              runMethods(notifier);
            }
          // PS: NullPointerException ? @SetupClass methods have to be static
          }).runProtected();

      fireAfterTest();
    }
  }

  @Override
  protected Object createTest() throws Exception
  {
    Object obj = currentTestData.createTest(getTestClass());
    if (obj != null && myCreateFixureMethod != null)
    {
      myCreateFixureMethod.invoke(obj);
    }
    if (obj != null)
    {
      fireCreateTestEvent(currentTestData, obj);
    }
    return obj;
  }

  private void fireAfterTest()
  {
    for (IParameterizedTestListener listener : getTestExecution())
    {
      try
      {
        listener.afterTestExecution();
      }
      catch (Exception e)
      {
        throw new RuntimeException(e);
      }
    }
  }

  private void fireCreateTestEvent(ITestData testData, Object testInstance)
      throws Exception
  {
    for (IParameterizedTestListener listener : getTestExecution())
    {
      listener.afterCreateTest(testInstance);
    }
  }

  protected Description methodDescription(Method method)
  {
    return currentTestData.getDescription();
  }

  /**
   * super.fTestMethods ist nicht initialisiert, da getTestMethods im
   * super-Konstruktor nur die
   *
   * @Test Annotation beachtet.
   */
  protected void runMethods(final RunNotifier notifier)
  {
    invokeTestMethod(myTestMethod, notifier);
  }

  /**
   * Damit der invoke mit Parameter aufgefllt wird, muss ein andere
   * TestMethod Wrapper erzeugt werden.
   */
  protected TestMethod wrapMethod(Method method)
  {
    return new TestDataMethod(method, myTestClass, currentTestData);
  }

  public ArrayList<IParameterizedTestListener> getTestExecution()
  {
    return myTestExecution;
  }
}

/**
 * Wrapper, that can invoke a test method with parameters.
 */
class TestDataMethod extends TestMethod
{

  /** The ITestData instance */
  private ITestData fTestData;

  /** Reference to the test method */
  private Method fMethod;

  /**
   * Constructs a TestDataMethod instance and setups the internal
   * fields to control to later pass them on the Test invocation
   *
   * @param testMethod
   * @param testClass
   * @param testData
   */
  public TestDataMethod(Method testMethod, TestClass testClass,
      ITestData testData)
  {
    super(testMethod, testClass);
    fTestData = testData;
    fMethod = testMethod;
  }

  /**
   * Invokes the test method on the given test class instance.
   *
   * @param testClassInstance a instance of the test class on which
   *        the test method resides.
   */
  public void invoke(Object testClassInstance) throws IllegalArgumentException,
      IllegalAccessException, InvocationTargetException
  {
    fTestData.invoke(fMethod, testClassInstance);
  }
}
