package com.seitenbau.junit;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.junit.internal.runners.TestClass;
import org.junit.runner.Description;

import com.seitenbau.junit.annotation.ParameterizedTestField;
import com.seitenbau.junit.model.DDTModelRoot;
import com.seitenbau.junit.model.DDTModelTest;
import com.seitenbau.junit.model.DDTModelTestExpectedResult;

/**
 * Default Implementation for invoking a test method with parameters.
 *
 */
public class TestData implements ITestData
{

  private Class<?> myTestClass;

  private int mySequenceNumber;

  private Method myMethod;

  private IDataTestProvider myProvider;

  private Properties myFieldData = new Properties();

  private Object[] myParameters;

  private ITestContext myTestContext;

  private String myDescriptionPostText;

  private boolean myPassContext;

  private ITestContext myContext;

  private DDTModelRoot myModel;

  private DDTModelTest myTestItem;

  public TestData(IDataTestProvider provider, int iSequenceNumber,
      DDTModelRoot root, DDTModelTest testItem)
  {
    if (provider == null)
    {
      throw new IllegalArgumentException();
    }
    init(provider, root, testItem, iSequenceNumber);
  }

  private void init(IDataTestProvider provider, DDTModelRoot root,
      DDTModelTest testItem, int sequenceNumber)
  {
    if (testItem == null)
    {
      throw new IllegalArgumentException();
    }

    myProvider = provider;
    myMethod = provider.getTestMethod();
    mySequenceNumber = sequenceNumber;
    myTestClass = provider.getTestClass();
    myModel = root;
    myTestItem = testItem;

    Method method = provider.getTestMethod();
    String name = method.getName();
    if (testItem.getName() != null)
    {
      name = testItem.getName() + " [" + method.getName() + "]";
    }
    if (testItem.getId() != null)
    {
      name += " [" + testItem.getId() + "]";
    }

    setDescriptionPostText(name);

    if (myTestItem.getParameters() != null
        && !myTestItem.getParameters().isEmpty())
    {
      if (myTestItem.getExpected() != null)
      {
        setParameters(myTestItem.getExpected().getExpectedResult(), myTestItem
            .getParameters().toArray());
      }
      else
      {
        setParameters(myTestItem.getParameters().toArray());
      }
    }
    else if (myTestItem.getExpected() != null)
    {
      setParameters(new Object[] {myTestItem.getExpected().getExpectedResult()});
    }
  }

  public ITestContext getContext()
  {
    return myContext;
  }

  public void setContext(ITestContext context)
  {
    getFieldData().put(ParameterizedTestField.DEFAULT_TESTCONTEXT_KEY, context);
    myContext = context;
  }

  public boolean isPassContext()
  {
    return myPassContext;
  }

  public void setPassContext(boolean passContext)
  {
    myPassContext = passContext;
  }

  public Description getDescription()
  {
    return Description.createTestDescription(myTestClass, getDescriptionText());
  }

  public Object[] getParameters()
  {

     return myParameters;

  }

  public Object createTest(TestClass testClass)
  {
    myContext.setTestItem(myTestItem);
    getFieldData().put(ParameterizedTestField.DEFAULT_TESTCONTEXT_KEY, myContext);
    Object obj;
    try
    {
      obj = testClass.getConstructor().newInstance();
    }
    catch (Exception e)
    {
      throw new RuntimeException(e);
    }
    try
    {
      // Setting up fields
      setupTest(obj);
    }
    catch (IllegalAccessException e)
    {
      e.printStackTrace();
    }
    return obj;
  }

  protected void setupTest(Object instance) throws IllegalAccessException
  {
    if (instance != null)
    {
      List<Field> fields = getAnnotatedFields(ParameterizedTestField.class);
      for (Field field : fields)
      {
        ParameterizedTestField annot = field
            .getAnnotation(ParameterizedTestField.class);
        String propKey = annot.propertyID();
        if (propKey != null)
        {
          Object value = getFieldData().get(propKey);
          field.setAccessible(true);

          try
          {
            field.getType().cast(value);
            field.set(instance, value);
          }
          catch (ClassCastException e)
          {
            throw new RuntimeException("Field " + field.getName()
                + " is not assignable by type " + value.getClass(), e);
          }
        }
      }
    }
  }

  protected List<Field> getAnnotatedFields(
      Class<? extends Annotation> annotationClass)
  {
    List<Field> results = new ArrayList<Field>();
    Field[] fields = myTestClass.getDeclaredFields();
    for (Field eachField : fields)
    {
      Annotation annotation = eachField.getAnnotation(annotationClass);
      if (annotation != null)
      {
        results.add(eachField);
      }
    }
    return results;
  }

  public int getSequenceNumber()
  {
    return mySequenceNumber;
  }

  public void setSequenceNumber(int sequenceNumber)
  {
    mySequenceNumber = sequenceNumber;
  }

  public Method getMethod()
  {
    return myMethod;
  }

  public void setMethod(Method method)
  {
    myMethod = method;
  }

  public IDataTestProvider getProvider()
  {
    return myProvider;
  }

  public void setProvider(IDataTestProvider provider)
  {
    myProvider = provider;
  }

  public Class<?> getTestClass()
  {
    return myTestClass;
  }

  public void setTestClass(Class<?> testClass)
  {
    myTestClass = testClass;
  }

  public ITestContext getTestContext()
  {
    return myTestContext;
  }

  public void setTestContext(ITestContext testContext)
  {
    getFieldData().put("TestContext", testContext);
    myTestContext = testContext;
  }

  public void setParameters(Object expectedResult, Object[] array)
  {
    int len = array.length;
    Object[] params = new Object[len + 1];
    params[0] = expectedResult;
    System.arraycopy(array, 0, params, 1, len);
    setParameters(params);
  }

  public void setParameters(Object[] parameters)
  {
    myParameters = parameters;
  }

  public void invoke(Method method, Object testInstance)
      throws IllegalArgumentException, IllegalAccessException,
      InvocationTargetException
  {
    List<Object> parameters = buildParameters();

    method.invoke(testInstance , parameters.toArray() );
  }

  protected List<Object> buildParameters()
  {
    DDTModelTest test = getTestItem();
    List<Object> parameters = new ArrayList<Object>();

    if(test.getExpected()!=null)
    {
      DDTModelTestExpectedResult expected = test.getExpected();
      if( expected.getExpectedResult()!=null)
      {
        parameters.add( expected.getExpectedResult() );
      }
    }


    buildRealParameters(parameters);

    if (isPassContext())
    {
      parameters.add(getContext());
    }
    return parameters;
  }

  protected void buildRealParameters(List<Object> parameters)
  {
    List<Object> realParameters = getTestItem().getParameters();
    if(realParameters!=null)
    {
      parameters.addAll( realParameters );
    }
  }

  protected String getDescriptionText()
  {
    return getSequenceNumber() + ": " + myDescriptionPostText;
  }

  public String getDescriptionPostText()
  {
    return myDescriptionPostText;
  }

  public void setDescriptionPostText(String descriptionPostText)
  {
    myDescriptionPostText = descriptionPostText;
  }

  public Properties getFieldData()
  {
    if (myFieldData == null)
    {
      myFieldData = new Properties();
    }
    return myFieldData;
  }

  public void setFieldData(Properties fieldData)
  {
    myFieldData = fieldData;
  }

  public DDTModelRoot getModel()
  {
    return myModel;
  }

  public void setModel(DDTModelRoot model)
  {
    myModel = model;
  }

  public DDTModelTest getTestItem()
  {
    return myTestItem;
  }

  public void setTestItem(DDTModelTest testItem)
  {
    myTestItem = testItem;
  }
}
