package com.seitenbau.junit;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import com.seitenbau.junit.annotation.XmlDataOption;
import com.seitenbau.junit.model.DDTModelRoot;
import com.seitenbau.junit.model.DDTModelTest;
import com.seitenbau.junit.model.DDTModelTestExpectedResult;
import com.thoughtworks.xstream.XStream;

/**
 * DataProvider implementation for XML based Data files.
 * 
 * This uses XStream to read the XML files. Therefore also non
 * primitive Data is allowed. See DDTModelRoot, which is used by
 * XStrem as base for the XML file.
 * 
 */
public class XmlDataProvider extends AbstractProvider
{

  /**
   * The xml file with the test Data. Must be deserializable into a
   * DDTModelRoot by xstream
   */
  private String myFileName;

  /**
   * Read in the XML file an build up the model
   */
  public Collection<? extends ITestData> buildTestData()
  {
    XmlDataOption annotation = getTestMethod().getAnnotation(
        XmlDataOption.class);

    if (annotation == null)
    {
      throw new IllegalArgumentException("Annotation ["
          + XmlDataOption.class.getCanonicalName() + "] is missing.");
    }

    setFileName(annotation.file());

    DDTModelRoot fileModel = loadModelFromFile(getFileName());

    boolean passContext = annotation.passContext();

    List<ITestData> data = buildTestDataFromModel(fileModel, passContext);

    return data;
  }

  /**
   * does the actual deserialization via XStream
   * 
   * @param fileName
   * 
   * @return
   */
  private DDTModelRoot loadModelFromFile(String fileName)
  {
    Class<?> testClass = getTestClass();
    if (testClass == null)
    {
      throw new IllegalArgumentException("TestClass was null");
    }
    InputStream stream = testClass.getResourceAsStream(fileName);
    if (stream == null)
    {
      throw new IllegalArgumentException("Unable to locate XML file: "
          + fileName + " beside Class: " + testClass.getCanonicalName());
    }

    XStream xs = getXStream();
    Object obj = xs.fromXML(stream);

    if (!(obj instanceof DDTModelRoot))
    {
      throw new IllegalArgumentException("Deserialized Object [file="
          + fileName + "] was not an instance of : "
          + DDTModelRoot.class.getCanonicalName());
    }

    return (DDTModelRoot) obj;
  }

  /**
   * Builds up the ITestData Collection for the passed model
   * 
   * @param modelRoot
   * @param passContextAsParameter
   * @return
   */
  private List<ITestData> buildTestDataFromModel(DDTModelRoot modelRoot,
      boolean passContextAsParameter)
  {
    List<ITestData> data = new ArrayList<ITestData>();
    int sequenceNumber = 0;
    for (DDTModelTest testModel : modelRoot.getTests())
    {
      TestData testData = createTestData(sequenceNumber++, modelRoot, testModel);

      setupTestData(testData, modelRoot, testModel, passContextAsParameter,
          sequenceNumber);

      data.add(testData);
    }
    return data;
  }

  protected void setupTestData(TestData testData, DDTModelRoot testModelRoot,
      DDTModelTest testItem, boolean passContextAsParameter, int sequenceNumber)
  {
    testData.setFieldData(testModelRoot.getProperties());
    testData.setPassContext(passContextAsParameter);

    setupTestDataContext(testData, testModelRoot, testItem, sequenceNumber);
  }

  protected void setupTestDataContext(TestData testData, DDTModelRoot testFile,
      DDTModelTest testItem, int iSeqNr)
  {
    ITestContext context = createContext(iSeqNr, testFile, testItem);
    context.setTestClass(getTestClass());
    testData.setContext(context);
  }

  /**
   * Create the concrete TestData instance for with the passed
   * Arguments
   * @param sequenceNumber
   * @param testfile
   * @param testItem
   * @return
   */
  protected TestData createTestData(int sequenceNumber, DDTModelRoot testfile,
      DDTModelTest testItem)
  {
    return new TestData(this, sequenceNumber, testfile, testItem);
  }

  /**
   * Create the concrete ITestContext instance for the given
   * Parameters
   * @param seqNr
   * @param model
   * @param testItem
   * @return
   */
  protected ITestContext createContext(int seqNr, DDTModelRoot model,
      DDTModelTest testItem)
  {
    return new XmlDataContext(this, model, testItem);
  }

  /**
   * 
   * @return the XStream instance that can serialize/deserialize the
   *         XML file
   */
  protected XStream getXStream()
  {
    XStream xs = new XStream();
    setupXStreamConverters(xs);
    setupXStreamFields(xs);
    return xs;
  }

  /**
   * fine tune the created XStream instance
   * @param xs
   */
  protected void setupXStreamFields(XStream xs)
  {
    xs.alias("testfile", DDTModelRoot.class);
    xs.alias("test", DDTModelTest.class);

    xs.useAttributeFor(DDTModelTest.class, "name");
    xs.useAttributeFor(DDTModelTest.class, "id");

    xs.useAttributeFor(DDTModelTestExpectedResult.class, "isException");
  }

  /**
   * fine tune the created XStream instance
   * @param xs
   */
  protected void setupXStreamConverters(XStream xs)
  {
  }

  /**
   * @return provides the post text for the JUnit GUI
   */
  public String getDescriptionPostText()
  {
    return myFileName;
  }

  /**
   * @return The XML filename which contains the XML data
   */
  public String getFileName()
  {
    return myFileName;
  }

  /**
   * Set the XML filename which contains the XML data
   */
  public void setFileName(String fileName)
  {
    myFileName = fileName;
  }
}
